package metrics;

import java.io.PrintWriter;

import AST.ASTNode;

/**
 * A metric to be calculated for nodes of the given node type <b>I</b> 
 *
 * @param <I> the node type this metric can be calculated for
 */
public interface Metric<I> {
	/**
	 * Returns the name of the metric to be used for output file names and internal
	 * identification of the metric.
	 * 
	 * @return the metric's name
	 */
	public String getName();
	
	/**
	 * Invoked before the first metric is to be collected so that the header
	 * of the output file is set up by the metric.
	 * 
	 * @param writer to write the header to
	 */
	public void setupHeader(PrintWriter writer);
	
	/**
	 * Calculate the metric for the given node. The {@link ASTNode#logMetric(String, String...)}
	 * method shall be used to output the results of the calculation.
	 * 
	 * @param node the node to calculate the metric for
	 * @see ASTNode#logMetric(String, String...)
	 */
	public void calculate(I node);
}
