package tests;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;

class Logger {
	private static final HashMap<String,PrintWriter> writers = new HashMap<String,PrintWriter>();
	
	/**
	 * Creates a new Logger that will create PrintWriters on demand and will close
	 * them upon system exit
	 */
	public Logger() {
		Runtime.getRuntime().addShutdownHook(new Thread() {
			@Override
			public void run() {
				System.err.println("Closing log writers ...");
				for(PrintWriter writer : writers.values()) {
					writer.close();
				}
			}});
	}
	
	public PrintWriter getPrintWriter(String fileName) throws IOException {
		PrintWriter writer = null;
		if(writers.containsKey(fileName)) {
		   writer = writers.get(fileName);
		} else {
			writer = new PrintWriter(new BufferedWriter(new FileWriter("foo.out")));
			writers.put(fileName, writer);
		}
		return writer;
	}
	
	public void log(String fileName, String line) {
		try {
			getPrintWriter(fileName).println(line);
		} catch (IOException ioe) {
			throw new RuntimeException(ioe);
		}
	}
}

class Bag {

}